# -*- coding: utf-8 -*-
#    Copyright (C) 2011 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

from __future__ import division
import os.path

import sip
from .. import qtall as qt4
from .. import utils
from .. import setting

def _(text, disambiguation=None, context="Tutorial"):
    """Translate text."""
    return qt4.QCoreApplication.translate(context, text, disambiguation)

class TutorialStep(qt4.QObject):

    nextStep = qt4.pyqtSignal()

    def __init__(self, text, mainwin,
                 nextstep=None, flash=None,
                 disablenext=False,
                 closestep=False,
                 nextonsetting=None,
                 nextonselected=None):

        """
        nextstep is class next TutorialStep class to use
        If flash is set, flash widget
        disablenext: wait until nextStep is emitted before going to next slide
        closestep: add a close button
        nextonsetting: (setnpath, lambda val: ok) -
          check setting to go to next slide
        nextonselected: go to next if widget with name is selected
        """

        qt4.QObject.__init__(self)
        self.text = text
        self.nextstep = nextstep
        self.flash = flash
        self.disablenext = disablenext
        self.closestep = closestep
        self.mainwin = mainwin

        self.nextonsetting = nextonsetting
        if nextonsetting is not None:
            mainwin.document.signalModified.connect(self.slotNextSetting)

        self.nextonselected = nextonselected
        if nextonselected is not None:
            mainwin.treeedit.widgetsSelected.connect(self.slotWidgetsSelected)

    def slotNextSetting(self, *args):
        """Check setting to emit next."""
        try:
            setn = self.mainwin.document.resolveSettingPath(
                None, self.nextonsetting[0]).get()
            if self.nextonsetting[1](setn):
                self.nextStep.emit()
        except ValueError:
            pass

    def slotWidgetsSelected(self, widgets, *args):
        """Go to next page if widget selected."""
        if len(widgets) == 1 and widgets[0].name == self.nextonselected:
            self.nextStep.emit()

##########################
## Introduction to widgets

class StepIntro(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Welcome to Veusz!</h1>

<p>This tutorial aims to get you working with Veusz as quickly as
possible.</p>

<p>You can close this tutorial at any time using the close button to
the top-right of this panel. The tutorial can be replayed in the help
menu.</p>

<p class="usercmd">Press Next to go to the next step</p>
'''), mainwin, nextstep=StepWidgets1)

class StepWidgets1(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Widgets</h1>

<p>Plots in Veusz are constructed from <i>widgets</i>.  Different
types of widgets are used to make different parts of a plot. For
example, there are widgets for axes, for a graph, for plotting data
and for plotting functions.</p>

<p>There are also special widgets. The grid widget arranges graphs
inside it in a grid arrangement.</p>
'''), mainwin, nextstep=StepWidgets2)

class StepWidgets2(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>Widgets can often be placed inside each other. For instance, a graph
widget is placed in a page widget or a grid widget. Plotting widgets
are placed in graph widget.</p>

<p>You can have multiple widgets of different types. For example, you
can have several graphs on the page, optionally arranged in a
grid. Several plotting widgets and axis widgets can be put in a
graph.</p>
'''), mainwin, nextstep=StepWidgetWin)

class StepWidgetWin(TutorialStep):
    def __init__(self, mainwin):
        t = mainwin.treeedit
        TutorialStep.__init__(
            self, _('''
<h1>Widget editing</h1>

<p>The flashing window is the Editing window, which shows the widgets
currently in the plot in a hierarchical tree. Each widget has a name
(the left column) and a type (the right column).</p>

<p class="usercmd">Press Next to continue.</p>
'''), mainwin,
            nextstep=StepWidgetWinExpand,
            flash=t)

class StepWidgetWinExpand(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>The graph widget is the currently selected widget.</p>

<p class="usercmd">Expand the graph widget - click the arrow or plus
to its left in the editing window - and select the x axis widget.</p>
'''), mainwin,
            disablenext=True,
            nextonselected='x',
            nextstep=StepPropertiesWin)

class StepPropertiesWin(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Widget properties</h1>

<p>This window shows the properties of the currently selected widget,
the x axis widget of the graph.</p>

<p class="usercmd">Enter a new label for the widget, by clicking in the
text edit box to the right of "Label", typing some text and press the
Enter key.</p>
'''), mainwin,
            flash = mainwin.propdock,
            disablenext = True,
            nextonsetting = ('/page1/graph1/x/label',
                             lambda val: val != ''),
            nextstep = StepPropertiesWin2)

class StepPropertiesWin2(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>Notice that the x axis label of your plot has now been updated.
Veusz supports LaTeX style formatting for labels, so you could include
superscripts, subscripts and fractions.</p>

<p>Other important axis properties include the minimum, maximum values
of the axis and whether the axis is logarithmic.</p>

<p class="usercmd">Click Next to continue.</p>
'''), mainwin, nextstep=WidgetAdd)

class WidgetAdd(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Adding widgets</h1>

<p>The flashing Add Widget toolbar and the Insert menu add widgets to
the document. New widgets are inserted in the currently selected
widget, if possible, or its parents.</p>

<p>Hold your mouse pointer over one of the toolbar buttons to
see a description of a widget type.</p>

<p class="usercmd">Press Next to continue.</p>
'''), mainwin,
            flash=mainwin.treeedit.addtoolbar,
            nextstep=FunctionAdd )

class FunctionAdd(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Add a function</h1>

<p>We will now add a function plotting widget to the current
graph.</p>

<p class="usercmd">Click on the flashing icon, or go to the Insert menu
and choosing "Add function".</p>
'''), mainwin,
            flash=mainwin.treeedit.addtoolbar.widgetForAction(
                mainwin.vzactions['add.function']),
            disablenext=True,
            nextonsetting = ('/page1/graph1/function1/function',
                             lambda val: val != ''),
            nextstep=FunctionSet)

class FunctionSet(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>You have now added a function widget to the graph widget. By
default function widgets plot y=x.</p>

<p class="usercmd">Go to the Function property and change the function to
be <code>x**2</code>, plotting x squared.</p>

<p>(Veusz uses Python syntax for its functions, so the power operator
is <code>**</code>, rather than <code>^</code>)</p>
'''), mainwin,
            nextonsetting = ('/page1/graph1/function1/function',
                             lambda val: val.strip() == 'x**2'),
            disablenext = True,
            nextstep=FunctionFormatting)

class FunctionFormatting(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Formatting</h1>

<p>Widgets have a number of formatting options. The Formatting window
(flashing) shows the options for the currently selected widget, here
the function widget.</p>

<p class="usercmd">Press Next to continue</p>
'''), mainwin,
            flash=mainwin.formatdock,
            nextstep=FunctionFormatLine)

class FunctionFormatLine(TutorialStep):
    def __init__(self, mainwin):

        tb = mainwin.formatdock.tabwidget.tabBar()
        label = qt4.QLabel("  ", tb)
        tb.setTabButton(1, qt4.QTabBar.LeftSide, label)

        TutorialStep.__init__(
            self, _('''
<p>Different types of formatting properties are grouped under separate
tables. The options for drawing the function line are grouped under
the flashing Line tab (%s).</p>

<p class="usercmd">Click on the Line tab to continue.</p>
''') % utils.pixmapAsHtml(utils.getPixmap('settings_plotline.png')),
            mainwin,
            flash=label,
            disablenext=True,
            nextstep=FunctionLineFormatting)

        tb.currentChanged[int].connect(self.slotCurrentChanged)

    def slotCurrentChanged(self, idx):
        if idx == 1:
            self.nextStep.emit()

class FunctionLineFormatting(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>Veusz lets you choose a line style, thickness and color for the
function line.</p>

<p class="usercmd">Choose a new line color for the line.</p>
'''),
            mainwin,
            disablenext=True,
            nextonsetting = ('/page1/graph1/function1/Line/color',
                             lambda val: val.strip() != 'black'),
            nextstep=DataStart)

###########
## Datasets

class DataStart(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Datasets</h1>

<p>Many widgets in Veusz plot datasets. Datasets can be imported from
files, entered manually or created from existing datasets using
operations or expressions.</p>

<p>Imported data can be linked to an external file or embedded in the
document.</p>

<p class="usercmd">Press Next to continue</p>
'''), mainwin,
            nextstep=DataImport)

class DataImport(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Importing data</h1>

<p>Let us start by importing data.</p>

<p class="usercmd">Click the flashing Data Import icon, or choose
"Import..."  From the Data menu.</p>
'''), mainwin,
            flash=mainwin.datatoolbar.widgetForAction(
                mainwin.vzactions['data.import']),
            disablenext=True,
            nextstep=DataImportDialog)

        # make sure we have the default delimiters
        for k in ( 'importdialog_csvdelimitercombo_HistoryCombo',
                   'importdialog_csvtextdelimitercombo_HistoryCombo' ):
            if k in setting.settingdb:
                del setting.settingdb[k]

        mainwin.dialogShown.connect(self.slotDialogShown)

    def slotDialogShown(self, dialog):
        """Called when a dialog is opened in the main window."""
        from ..dialogs.importdialog import ImportDialog
        if isinstance(dialog, ImportDialog):
            # make life easy by sticking in filename
            dialog.slotReset()
            dialog.filenameedit.setText(
                os.path.join(utils.exampleDirectory, 'tutorialdata.csv'))
            # and choosing tab
            dialog.guessImportTab()
            # get rid of existing values
            self.nextStep.emit()

class DataImportDialog(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>This is the data import dialog. In this tutorial, we have selected
an example CSV (comma separated value) file for you, but you would
normally browse to find your data file.</p>

<p>This example file defines three datasets, <i>alpha</i>, <i>beta</i>
and <i>gamma</i>, entered as columns in the CSV file.</p>

<p class="usercmd">Press Next to continue</p>
'''), mainwin, nextstep=DataImportDialog2)

class DataImportDialog2(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>Veusz will try to guess the datatype - numeric, text or date - from
the data in the file or you can specify it manually.</p>

<p>Several different data formats are supported in Veusz and plugins
can be defined to import any data format. The Link option links data
to the original file.</p>

<p class="usercmd">Click the Import button in the dialog.</p>
'''), mainwin,
            nextstep=DataImportDialog3,
            disablenext=True)
        mainwin.document.signalModified.connect(self.slotDocModified)

    def slotDocModified(self):
        if 'alpha' in self.mainwin.document.data:
            self.nextStep.emit()

class DataImportDialog3(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>Notice how Veusz has loaded the three different datasets from the
file. You could carry on importing new datasets from the Import dialog
box or reopen it later.</p>

<p class="usercmd">Close the Import dialog box.</p>
'''), mainwin,
            disablenext=True,
            nextstep=DataImportDialog4)

        self.timer = qt4.QTimer()
        self.timer.timeout.connect(self.slotTimeout)
        self.timer.start(200)

    def slotTimeout(self):
        from ..dialogs.importdialog import ImportDialog
        closed = True
        for dialog in self.mainwin.dialogs:
            if isinstance(dialog, ImportDialog):
                closed = False
        if closed:
            # move forward if no import dialog open
            self.nextStep.emit()

class DataImportDialog4(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>The Data viewing window (flashing) shows the currently loaded
datasets in the document.</p>

<p>Hover your mouse over datasets to get information about them. You
can see datasets in more detail in the Data Edit dialog box.</p>

<p class="usercmd">Click Next to continue</p>
'''), mainwin,
            flash=mainwin.datadock,
            nextstep=AddXYPlotter)

##############
## XY plotting

class AddXYPlotter(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Plotting data</h1>

<p>The point plotting widget plots datasets loaded in Veusz.</p>

<p class="usercmd">The flashing icon adds a point plotting (xy)
widget. Click on this, or go to the Insert menu and choose "Add xy".</p>
'''), mainwin,
            flash=mainwin.treeedit.addtoolbar.widgetForAction(
                mainwin.vzactions['add.xy']),
            disablenext=True,
            nextonsetting = ('/page1/graph1/xy1/xData',
                             lambda val: val != ''),
            nextstep=SetXY_X)

class SetXY_X(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>The datasets to be plotted are in the widget's properties.</p>

<p class="usercmd">Change the "X data" setting to be the
<code>alpha</code> dataset. You can choose this from the drop down
menu or type it.</p>
'''), mainwin,
            disablenext=True,
            nextonsetting = ('/page1/graph1/xy1/xData',
                             lambda val: val == 'alpha'),
            nextstep=SetXY_Y)

class SetXY_Y(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p class="usercmd">Change the "Y data" setting to be the
<code>beta</code> dataset.</p>
'''), mainwin,
            disablenext=True,
            nextonsetting = ('/page1/graph1/xy1/yData',
                             lambda val: val == 'beta'),
            nextstep=SetXYLine)

class SetXYLine(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>Veusz has now plotted the data on the graph. You can manipulate how
the data are shown using the formatting settings.</p>

<p class="usercmd">Make sure that the line Formatting tab (%s) for the
widget is selected.</p>

<p class="usercmd">Click on the check box next to the Hide option at
the bottom, to hide the line plotted between the data points.</p>
''') % utils.pixmapAsHtml(utils.getPixmap('settings_plotline.png')),
            mainwin,
            disablenext=True,
            nextonsetting = ('/page1/graph1/xy1/PlotLine/hide',
                             lambda val: val),
            nextstep=SetXYFill)

class SetXYFill(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>Now we will change the point color.</p>

<p class="usercmd">Click on the "Marker fill (%s)" formatting tab.
Change the fill color of the plotted data.</p>
''') % utils.pixmapAsHtml(utils.getPixmap('settings_plotmarkerfill.png')),
            mainwin,
            disablenext=True,
            nextonsetting = ('/page1/graph1/xy1/MarkerFill/color',
                             lambda val: val != 'black'),
            nextstep=AddXY2nd)

class AddXY2nd(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Adding a second dataset</h1>

<p>We will now plot dataset <code>alpha</code> against
<code>gamma</code> on the same graph.</p>

<p class="usercmd">Add a second point plotting (xy) widget using the
flashing icon, or go to the Insert menu and choose "Add xy".</p>
'''), mainwin,
            flash=mainwin.treeedit.addtoolbar.widgetForAction(
                mainwin.vzactions['add.xy']),
            disablenext=True,
            nextonsetting = ('/page1/graph1/xy2/xData',
                             lambda val: val != ''),
            nextstep=AddXY2nd_2)

class AddXY2nd_2(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p class="usercmd">Change the "X data" setting to be the
<code>alpha</code> dataset.</p>
'''), mainwin,
            disablenext=True,
            nextonsetting = ('/page1/graph1/xy2/xData',
                             lambda val: val == 'alpha'),
            nextstep=AddXY2nd_3)

class AddXY2nd_3(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p class="usercmd">Next, change the "Y data" setting to be the
<code>gamma</code> dataset.</p>
'''), mainwin,
            disablenext=True,
            nextonsetting = ('/page1/graph1/xy2/yData',
                             lambda val: val == 'gamma'),
            nextstep=AddXY2nd_4)

class AddXY2nd_4(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>We can fill regions under plots using the Fill Below Formatting tab
(%s).</p>

<p class="usercmd">Go to this tab, and unselect the "Hide edge fill"
option.</p>
''') % utils.pixmapAsHtml(utils.getPixmap('settings_plotfillbelow.png')),
            mainwin,
            disablenext=True,
            nextonsetting = ('/page1/graph1/xy2/FillBelow/hide',
                             lambda val: not val),
            nextstep=File1)

class File1(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Saving</h1>

<p>The document can be saved under the File menu, choosing "Save
as...", or by clicking on the Save icon (flashing).</p>

<p>Veusz documents are simple text files which can be easily modified
outside the program.</p>

<p class="usercmd">Click Next to continue</p>
'''), mainwin,
            flash=mainwin.maintoolbar.widgetForAction(
                mainwin.vzactions['file.save']),
            nextstep=File2)

class File2(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Exporting</h1>

<p>The document can be exported in scalable (EPS, PDF, SVG and EMF) or
bitmap formats.</p>

<p>The "Export..." command under the File menu exports the selected
page. Alternatively, click on the Export icon (flashing).</p>

<p class="usercmd">Click Next to continue</p>
'''), mainwin,
            flash=mainwin.maintoolbar.widgetForAction(
                mainwin.vzactions['file.export']),
            nextstep=Cut1,
            )

class Cut1(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Cut and paste</h1>

<p>Widgets can be cut and pasted to manipulate the document.</p>

<p class="usercmd">Select the "graph1" widget in the Editing window.</p>
'''), mainwin,
            disablenext=True,
            nextonselected='graph1',
            nextstep=Cut2)

class Cut2(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p class="usercmd">Now click the Cut icon (flashing) or choose "Cut"
from the Edit menu.</p>

<p>This copies the currently selected widget to the clipboard and
deletes it from the document.</p>
'''), mainwin,
            disablenext=True,
            flash=mainwin.treeedit.edittoolbar.widgetForAction(
                mainwin.vzactions['edit.cut']),
            nextstep=AddGrid)
        mainwin.document.signalModified.connect(self.slotCheckDelete)

    def slotCheckDelete(self, *args):
        d = self.mainwin.document
        try:
            d.resolveWidgetPath(None, '/page1/graph1')
        except ValueError:
            # success!
            self.nextStep.emit()

class AddGrid(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>Adding a grid</h1>

<p>Now we will add a grid widget to paste the graph back into.</p>

<p class="usercmd">Click on the flashing Grid widget icon, or choose
"Add grid" from the Insert menu.</p>
'''), mainwin,
            flash=mainwin.treeedit.addtoolbar.widgetForAction(
                mainwin.vzactions['add.grid']),
            disablenext=True,
            nextonsetting = ('/page1/grid1/rows',
                             lambda val: val != ''),
            nextstep=Paste1)

class Paste1(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p class="usercmd">Now click the Paste icon (flashing) or choose "Paste"
from the Edit menu.</p>

<p>This pastes back the widget from the clipboard.</p>
'''), mainwin,
            disablenext=True,
            flash=mainwin.treeedit.edittoolbar.widgetForAction(
                mainwin.vzactions['edit.paste']),
            nextonsetting = ('/page1/grid1/graph1/leftMargin',
                             lambda val: val != ''),
            nextstep=Paste2)

class Paste2(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p class="usercmd">For a second time, click the Paste icon (flashing)
or choose "Paste" from the Edit menu.</p>

<p>This adds a second copy of the original graph to the grid.</p>
'''), mainwin,
            disablenext=True,
            flash=mainwin.treeedit.edittoolbar.widgetForAction(
                mainwin.vzactions['edit.paste']),
            nextonsetting = ('/page1/grid1/graph2/leftMargin',
                             lambda val: val != ''),
            nextstep=Paste3)

class Paste3(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>Having the graphs side-by-side looks a bit messy. We would like to
change the graphs to be arranged in rows.</p>

<p class="usercmd">Navigate to the grid1 widget properties. Change the
number of columns to 1.</p>
'''), mainwin,
            disablenext=True,
            nextonsetting = ('/page1/grid1/columns',
                             lambda val: val == 1),
            nextstep=Paste4)

class Paste4(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<p>We could now adjust the margins of the graphs and the grid.</p>

<p>Axes can also be shared by the graphs of the grid by moving them
into the grid widget. This shares the same axis scale for graphs.</p>

<p class="usercmd">Click Next to continue</p>
'''), mainwin, nextstep=EndStep)

class EndStep(TutorialStep):
    def __init__(self, mainwin):
        TutorialStep.__init__(
            self, _('''
<h1>The End</h1>

<p>Thank you for working through this Veusz tutorial. We hope you
enjoy using Veusz!</p>

<p>Please send comments, bug reports and suggestions to the
developers via the mailing list.</p>

<p>You can try this tutorial again from the Help menu.</p>
'''), mainwin, closestep=True, disablenext=True)

class TutorialDock(qt4.QDockWidget):
    '''A dock tutorial window.'''

    def __init__(self, document, mainwin, *args):
        qt4.QDockWidget.__init__(self, *args)
        self.setAttribute(qt4.Qt.WA_DeleteOnClose)
        self.setMinimumHeight(300)
        self.setWindowTitle('Tutorial - Veusz')
        self.setObjectName('veusztutorialwindow')

        self.setStyleSheet('background: lightyellow;')

        self.document = document
        self.mainwin = mainwin

        self.layout = l = qt4.QVBoxLayout()

        txtdoc = qt4.QTextDocument(self)
        txtdoc.setDefaultStyleSheet(
            "p.usercmd { color: blue; } "
            "h1 { font-size: x-large;} "
            "code { color: green;} "
            )
        self.textedit = qt4.QTextEdit(readOnly=True)
        self.textedit.setDocument(txtdoc)

        l.addWidget(self.textedit)

        self.buttonbox = qt4.QDialogButtonBox()
        self.nextb = self.buttonbox.addButton(
            'Next', qt4.QDialogButtonBox.ActionRole)
        self.nextb.clicked.connect(self.slotNext)

        l.addWidget(self.buttonbox)

        # have to use a separate widget as dialog already has layout
        self.widget = qt4.QWidget()
        self.widget.setLayout(l)
        self.setWidget(self.widget)

        # timer for controlling flashing
        self.flashtimer = qt4.QTimer(self)
        self.flashtimer.timeout.connect(self.slotFlashTimeout)
        self.flash = self.oldflash = None
        self.flashon = False
        self.flashct = 0
        self.flashtimer.start(500)

        self.changeStep(StepIntro)

    def ensureShowFlashWidgets(self):
        '''Ensure we can see the widgets flashing.'''
        w = self.flash
        while w is not None:
            w.show()
            w = w.parent()

    def changeStep(self, stepklass):
        '''Apply the next step.'''

        # this is the current text
        self.step = stepklass(self.mainwin)

        # listen to step for next step
        self.step.nextStep.connect(self.slotNext)

        # update text
        self.textedit.setHtml(self.step.text)

        # handle requests for flashing
        self.flashct = 20
        self.flashon = True
        self.flash = self.step.flash
        if self.flash is not None:
            self.ensureShowFlashWidgets()

        # enable/disable next button
        self.nextb.setEnabled(not self.step.disablenext)

        # add a close button if requested
        if self.step.closestep:
            closeb = self.buttonbox.addButton(
                'Close', qt4.QDialogButtonBox.ActionRole)
            closeb.clicked.connect(self.close)

    # work around C/C++ object deleted
    @qt4.pyqtSlot()
    def slotFlashTimeout(self):
        '''Handle flashing of UI components.'''

        # because we're flashing random UI components, the C++ object
        # might be deleted, so we have to check before doing things to
        # it: hence the sip.isdeleted

        if ( self.flash is not self.oldflash and self.oldflash is not None
             and not sip.isdeleted(self.oldflash) ):
            # clear any flashing on previous widget
            self.oldflash.setStyleSheet('')
            self.oldflash = None

        if self.flash is not None and not sip.isdeleted(self.flash):
            # set flash state and toggle variable
            if self.flashon:
                self.flash.setStyleSheet('background: yellow;')
            else:
                self.flash.setStyleSheet('')
            self.flashon = not self.flashon
            self.oldflash = self.flash

            # stop flashing after N iterations
            self.flashct -= 1
            if self.flashct == 0:
                self.flash = None

    def slotNext(self):
        """Move to the next page of the tutorial."""
        nextstepklass = self.step.nextstep
        if nextstepklass is not None:
            self.changeStep( nextstepklass )



# version.py
# return the version number

#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""
Return Veusz' version number
"""

from __future__ import division
import os.path
import sys
import datetime

from . import utilfuncs
from ..compat import curlrequest
from .. import qtall as qt

_errmsg = """Failed to find VERSION file.

This is probably because the resource files are not installed in the
python module directory. You may need to set the environment variable
VEUSZ_RESOURCE_DIR or add a "resources" symlink in the main veusz
module directory pointing to the directory where resources are
located. See INSTALL for details.
"""

_ver = None
def version():
    """Return the version number as a string."""

    global _ver
    if _ver:
        return _ver

    filename = os.path.join(utilfuncs.resourceDirectory, "VERSION")
    try:
        with open(filename) as f:
            _ver = f.readline().strip()
        return _ver
    except EnvironmentError:
        sys.stderr.write(_errmsg)
        sys.exit(1)

def versionToTuple(ver):
    """Convert version to tuple, e.g. '2.1.1' -> (2,1,1)."""
    return tuple([int(x) for x in ver.split('.')])

def latestVersion():
    """Get latest version of Veusz from website as string.

    Returns None if error
    """

    try:
        f = curlrequest.urlopen(
            'http://veusz.github.io/download/newest-version.html')
        p = f.read()
        f.close()

        latest = p.decode('ascii').strip()
        # check format
        intver = versionToTuple(latest)
    except Exception:
        return None

    return latest

class VersionCheckThread(qt.QThread):
    """Asynchronously check for new version, emitting signal if found."""

    newversion = qt.pyqtSignal(str)

    # minimum number of days to wait between checks
    mininterval = 7

    def run(self):
        # can't import setting above because of loops
        from .. import setting

        if ( disableVersionChecks or
             setting.settingdb['vercheck_disabled'] or
             not setting.settingdb['vercheck_asked_user']):
            return

        today = datetime.date.today()
        dayssincecheck = (
            today -
            datetime.date(*setting.settingdb['vercheck_last_done'])).days

        if dayssincecheck >= self.mininterval or dayssincecheck < 0:
            setting.settingdb['vercheck_last_done'] = (
                today.year, today.month, today.day)
            #print("doing check")
            latestver = latestVersion()
            if latestver:
                setting.settingdb['vercheck_latest'] = latestver

        thisver = version()
        latestver = setting.settingdb['vercheck_latest']
        #print('latest ver', latestver, thisver)

        # is newer version available?
        if ( latestver and
             versionToTuple(latestver) > versionToTuple(thisver)):
            self.newversion.emit(latestver)

# patch this to be True if you are packaging Veusz and want to disable
# version checks
disableVersionChecks=False



# plotwindow.py
# the main window for showing plots

#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

from __future__ import division
import sys
import traceback

from ..compat import crange
from .. import qtall as qt4
import numpy as N

from .. import setting
from ..dialogs import exceptiondialog
from .. import document
from .. import utils
from .. import widgets

def _(text, disambiguation=None, context='PlotWindow'):
    """Translate text."""
    return qt4.QCoreApplication.translate(context, text, disambiguation)

class PickerCrosshairItem( qt4.QGraphicsPathItem ):
    """The picker cross widget: it moves from point to point and curve to curve
       with the arrow keys, and hides itself when it looses focus"""
    def __init__(self, parent=None):
        path = qt4.QPainterPath()
        path.addRect(-4, -4, 8, 8)
        path.addRect(-5, -5, 10, 10)
        path.moveTo(-8, 0)
        path.lineTo(8, 0)
        path.moveTo(0, -8)
        path.lineTo(0, 8)

        qt4.QGraphicsPathItem.__init__(self, path, parent)
        self.setBrush(qt4.QBrush(qt4.Qt.black))
        self.setFlags(self.flags() | qt4.QGraphicsItem.ItemIsFocusable)

    def paint(self, painter, option, widget):
        """Override this to enforce the global antialiasing setting"""
        aa = setting.settingdb['plot_antialias']
        painter.save()
        painter.setRenderHint(qt4.QPainter.Antialiasing, aa)
        qt4.QGraphicsPathItem.paint(self, painter, option, widget)
        painter.restore()

    def focusOutEvent(self, event):
        qt4.QGraphicsPathItem.focusOutEvent(self, event)
        self.hide()

class RenderControl(qt4.QObject):
    """Object for rendering plots in a separate thread."""

    # emitted when new item on plot queue
    sigQueueChange = qt4.pyqtSignal(int)

    # when a rendering job is finished
    signalRenderFinished = qt4.pyqtSignal(
        int, qt4.QImage, document.PaintHelper)

    def __init__(self, plotwindow):
        """Start up numthreads rendering threads."""
        qt4.QObject.__init__(self)
        self.sem = qt4.QSemaphore()
        self.mutex = qt4.QMutex()
        self.threads = []
        self.exit = False
        self.latestjobs = []
        self.latestaddedjob = -1
        self.latestdrawnjob = -1
        self.plotwindow = plotwindow

        self.updateNumberThreads()

    def updateNumberThreads(self, num=None):
        """Changes the number of rendering threads."""
        if num is None:
            if qt4.QFontDatabase.supportsThreadedFontRendering():
                # use number of threads in preference
                num = setting.settingdb['plot_numthreads']
            else:
                # disable threads
                num = 0

        if self.threads:
            # delete old ones
            self.exit = True
            self.sem.release(len(self.threads))
            for t in self.threads:
                t.wait()
            del self.threads[:]
            self.exit = False

        # start new ones
        for i in crange(num):
            t = RenderThread(self)
            t.start()
            self.threads.append(t)

    def exitThreads(self):
        """Exit threads started."""
        self.updateNumberThreads(num=0)

    def processNextJob(self):
        """Take a job from the queue and process it.

        emits renderfinished(jobid, img, painthelper)
        when done, if job has not been superseded
        """

        self.mutex.lock()
        jobid, helper = self.latestjobs[-1]
        del self.latestjobs[-1]
        lastadded = self.latestaddedjob
        self.mutex.unlock()

        # don't process jobs which have been superseded
        if lastadded == jobid:
            img = qt4.QImage(
                int(helper.rawpagesize[0]), int(helper.rawpagesize[1]),
                qt4.QImage.Format_ARGB32_Premultiplied)
            img.fill( setting.settingdb.color('page').rgb() )

            painter = qt4.QPainter(img)
            aa = self.plotwindow.antialias
            painter.setRenderHint(qt4.QPainter.Antialiasing, aa)
            painter.setRenderHint(qt4.QPainter.TextAntialiasing, aa)
            helper.renderToPainter(painter)
            painter.end()

            self.mutex.lock()
            # just throw away result if it older than the latest one
            if jobid > self.latestdrawnjob:
                self.signalRenderFinished.emit(jobid, img, helper)
                self.latestdrawnjob = jobid
            self.mutex.unlock()

        # tell any listeners that a job has been processed
        self.sigQueueChange.emit(-1)

    def addJob(self, helper):
        """Process drawing job in PaintHelper given."""

        # indicate that there is a new item to be processed to listeners
        self.sigQueueChange.emit(1)

        # add the job to the queue
        self.mutex.lock()
        self.latestaddedjob += 1
        self.latestjobs.append( (self.latestaddedjob, helper) )
        self.mutex.unlock()

        if self.threads:
            # tell a thread to process job
            self.sem.release(1)
        else:
            # process job in current thread if multithreading disabled
            self.processNextJob()

class RenderThread( qt4.QThread ):
    """A thread for processing rendering jobs.
    This is controlled by a RenderControl object
    """

    def __init__(self, rendercontrol):
        qt4.QThread.__init__(self)
        self.rc = rendercontrol

    def run(self):
        """Repeat forever until told to exit.
        If it aquires 1 resource from the semaphore it will process
        the next job.
        """
        while True:
            # wait until we can aquire the resources
            self.rc.sem.acquire(1)
            if self.rc.exit:
                break
            try:
                self.rc.processNextJob()
            except Exception:
                sys.stderr.write(_("Error in rendering thread\n"))
                traceback.print_exc(file=sys.stderr)

class ControlGraphRoot(qt4.QGraphicsItem):
    """Control graph items are connected to this root item.
    We don't use a group here as it would swallow parent events."""
    def __init__(self):
        qt4.QGraphicsItem.__init__(self)
    def paint(self, painter, option, widget=None):
        pass
    def boundingRect(self):
        return qt4.QRectF()

class PlotWindow( qt4.QGraphicsView ):
    """Class to show the plot(s) in a scrollable window."""

    # emitted when new item on plot queue
    sigQueueChange = qt4.pyqtSignal(int)
    # on drawing a page
    sigUpdatePage = qt4.pyqtSignal(int)
    # point picked on plot
    sigPointPicked = qt4.pyqtSignal(object)
    # picker enabled
    sigPickerEnabled = qt4.pyqtSignal(bool)
    # axis values update from moving mouse
    sigAxisValuesFromMouse = qt4.pyqtSignal(dict)
    # gives widget clicked
    sigWidgetClicked = qt4.pyqtSignal(object, str)

    # how often the document can update
    updateintervals = (
        (0, _('Disable')),
        (-1, _('On document change')),
        (100, _('Every 0.1s')),
        (250, _('Every 0.25s')),
        (500, _('Every 0.5s')),
        (1000, _('Every 1s')),
        (2000, _('Every 2s')),
        (5000, _('Every 5s')),
        (10000, _('Every 10s')),
        )

    def __init__(self, document, parent, menu=None):
        """Initialise the window.

        menu gives a menu to add any menu items to
        """

        qt4.QGraphicsView.__init__(self, parent)
        self.setBackgroundRole(qt4.QPalette.Dark)
        self.scene = qt4.QGraphicsScene()
        self.setScene(self.scene)

        # this graphics scene item is the actual graph
        pixmap = qt4.QPixmap(1, 1)
        self.dpi = (pixmap.logicalDpiX(), pixmap.logicalDpiY())
        self.pixmapitem = self.scene.addPixmap(pixmap)

        # whether full screen mode
        self.isfullscreen = False

        # set to be parent's actions
        self.vzactions = None

        # for controlling plot elements
        self.controlgraphroot = ControlGraphRoot()
        self.scene.addItem(self.controlgraphroot)

        # zoom rectangle for zooming into graph (not shown normally)
        self.zoomrect = self.scene.addRect(
            0, 0, 100, 100, qt4.QPen(qt4.Qt.DotLine))
        self.zoomrect.setZValue(2.)
        self.zoomrect.hide()

        # picker graphicsitem for marking the picked point
        self.pickeritem = PickerCrosshairItem()
        self.scene.addItem(self.pickeritem)
        self.pickeritem.setZValue(2.)
        self.pickeritem.hide()

        # all the widgets that picker key-navigation might cycle through
        self.pickerwidgets = []

        # the picker state
        self.pickerinfo = widgets.PickInfo()

        # set up so if document is modified we are notified
        self.document = document
        self.docchangeset = -100
        self.oldpagenumber = -1
        self.document.signalModified.connect(self.slotDocModified)

        # state of last plot from painthelper
        self.painthelper = None

        self.lastwidgetsselected = []
        self.oldzoom = -1.
        self.zoomfactor = 1.
        self.pagenumber = 0
        self.ignoreclick = False

        # for rendering plots in separate threads
        self.rendercontrol = RenderControl(self)
        self.rendercontrol.signalRenderFinished.connect(
            self.slotRenderFinished)
        self.rendercontrol.sigQueueChange.connect(
            self.sigQueueChange)

        # mode for clicking
        self.clickmode = 'select'
        self.currentclickmode = None

        # wheel zooming/scrolling accumulator
        self.sumwheeldelta = 0

        # set up redrawing timer
        self.timer = qt4.QTimer(self)
        self.timer.timeout.connect(self.checkPlotUpdate)

        # for drag scrolling
        self.grabpos = None
        self.scrolltimer = qt4.QTimer(self)
        self.scrolltimer.setSingleShot(True)

        # for turning clicking into scrolling after a period
        self.scrolltimer.timeout.connect(self.slotBecomeScrollClick)

        # get plot view updating policy
        #  -1: update on document changes
        #   0: never update automatically
        #  >0: check for updates every x ms
        self.interval = setting.settingdb['plot_updatepolicy']

        # if using a time-based document update checking, start timer
        if self.interval > 0:
            self.timer.start(self.interval)

        # load antialias settings
        self.antialias = setting.settingdb['plot_antialias']

        # allow window to get focus, to allow context menu
        self.setFocusPolicy(qt4.Qt.StrongFocus)

        # get mouse move events if mouse is not pressed
        self.setMouseTracking(True)

        # create toolbar in main window (urgh)
        self.createToolbar(parent, menu)

    def hideEvent(self, event):
        """Window closing, so exit rendering threads."""
        self.rendercontrol.exitThreads()
        qt4.QGraphicsView.hideEvent(self, event)

    def sizeHint(self):
        """Return size hint for window."""
        p = self.pixmapitem.pixmap()
        if p.width() <= 1 and p.height() <= 1:
            # if the document has been uninitialized, get the doc size
            return qt4.QSize(*self.document.docSize())
        return p.size()

    def showToolbar(self, show=True):
        """Show or hide toolbar"""
        self.viewtoolbar.setVisible(show)

    def createToolbar(self, parent, menu=None):
        """Make a view toolbar, and optionally update menu."""

        self.viewtoolbar = qt4.QToolBar(_("View toolbar - Veusz"), parent)
        self.viewtoolbar.setObjectName('veuszviewtoolbar')
        iconsize = setting.settingdb['toolbar_size']
        self.viewtoolbar.setIconSize(qt4.QSize(iconsize, iconsize))
        self.viewtoolbar.hide()
        if parent:
            parent.addToolBar(qt4.Qt.TopToolBarArea, self.viewtoolbar)

        if parent and hasattr(parent, 'vzactions'):
            # share actions with parent if possible
            # as plot windows can be isolated from mainwindows, we need this
            self.vzactions = actions = parent.vzactions
        else:
            self.vzactions = actions = {}

        a = utils.makeAction
        actions.update({
                'view.zoomin':
                    a(self, _('Zoom into the plot'), _('Zoom &In'),
                      self.slotViewZoomIn,
                      icon='kde-zoom-in', key='Ctrl++'),
                'view.zoomout':
                    a(self, _('Zoom out of the plot'), _('Zoom &Out'),
                      self.slotViewZoomOut,
                      icon='kde-zoom-out', key='Ctrl+-'),
                'view.zoom11':
                    a(self, _('Restore plot to natural size'), _('Zoom 1:1'),
                      self.slotViewZoom11,
                      icon='kde-zoom-1-veuszedit', key='Ctrl+1'),
                'view.zoomwidth':
                    a(self, _('Zoom plot to show whole width'), _('Zoom to width'),
                      self.slotViewZoomWidth,
                      icon='kde-zoom-width-veuszedit'),
                'view.zoomheight':
                    a(self, _('Zoom plot to show whole height'), _('Zoom to height'),
                      self.slotViewZoomHeight,
                      icon='kde-zoom-height-veuszedit'),
                'view.zoompage':
                    a(self, _('Zoom plot to show whole page'), _('Zoom to page'),
                      self.slotViewZoomPage,
                      icon='kde-zoom-page-veuszedit'),
                'view.zoommenu':
                    a(self, _('Zoom functions menu'), _('Zoom'),
                      self.doZoomMenuButton,
                      icon='kde-zoom-veuszedit'),
                'view.prevpage':
                    a(self, _('Move to the previous page'), _('&Previous page'),
                      self.slotViewPreviousPage,
                      icon='kde-go-previous', key='Ctrl+PgUp'),
                'view.nextpage':
                    a(self, _('Move to the next page'), _('&Next page'),
                      self.slotViewNextPage,
                      icon='kde-go-next', key='Ctrl+PgDown'),
                'view.select':
                    a(self, _('Select items from the graph or scroll'),
                      _('Select items or scroll'),
                      None,
                      icon='kde-mouse-pointer'),
                'view.pick':
                    a(self, _('Read data points on the graph'),
                      _('Read data points'),
                      None,
                      icon='veusz-pick-data'),
                'view.zoomgraph':
                    a(self, _('Zoom into graph'), _('Zoom graph'),
                      None,
                      icon='veusz-zoom-graph'),
                'view.fullscreen':
                    a(self, _('View plot full screen'), _('Full screen'),
                      self.slotFullScreen,
                      icon='veusz-view-fullscreen', key='Ctrl+F11'),
                })

        if menu:
            # only construct menu if required
            menuitems = [
                ('view', '', [
                        'view.zoomin', 'view.zoomout',
                        'view.zoom11', 'view.zoomwidth',
                        'view.zoomheight', 'view.zoompage',
                        '',
                        'view.prevpage', 'view.nextpage',
                        'view.fullscreen',
                        '',
                        'view.select', 'view.pick', 'view.zoomgraph',
                        ]),
                ]
            utils.constructMenus(menu, {'view': menu}, menuitems,
                                 actions)

        # populate menu on zoom menu toolbar icon
        zoommenu = qt4.QMenu(self)
        zoomag = qt4.QActionGroup(self)
        for act in ('view.zoomin', 'view.zoomout', 'view.zoom11',
                    'view.zoomwidth', 'view.zoomheight', 'view.zoompage'):
            a = actions[act]
            zoommenu.addAction(a)
            zoomag.addAction(a)
            a.vzname = act
        actions['view.zoommenu'].setMenu(zoommenu)
        zoomag.triggered.connect(self.zoomActionTriggered)

        lastzoom = setting.settingdb.get('view_defaultzoom', 'view.zoompage')
        self.updateZoomMenuButton(actions[lastzoom])

        # add items to toolbar
        utils.addToolbarActions(self.viewtoolbar, actions,
                                ('view.prevpage', 'view.nextpage',
                                 'view.fullscreen',
                                 'view.select', 'view.pick',
                                 'view.zoomgraph', 'view.zoommenu'))

        # define action group for various different selection models
        grp = self.selectactiongrp = qt4.QActionGroup(self)
        grp.setExclusive(True)
        for a in ('view.select', 'view.pick', 'view.zoomgraph'):
            actions[a].setActionGroup(grp)
            actions[a].setCheckable(True)
        actions['view.select'].setChecked(True)
        grp.triggered.connect(self.slotSelectMode)

        return self.viewtoolbar

    def zoomActionTriggered(self, action):
        """Keep track of the last zoom action selected."""
        setting.settingdb['view_defaultzoom'] = action.vzname
        self.updateZoomMenuButton(action)

    def updateZoomMenuButton(self, action):
        """Make zoom button call default zoom action and change icon."""
        menuact = self.vzactions['view.zoommenu']
        setting.settingdb['view_defaultzoom'] = action.vzname
        menuact.setIcon( action.icon() )

    def doZoomMenuButton(self):
        """Select previous zoom option when clicking on zoom menu."""
        act = self.vzactions[setting.settingdb['view_defaultzoom']]
        act.trigger()

    def doZoomRect(self, endpos):
        """Take the zoom rectangle drawn by the user and do the zooming.
        endpos is a QPoint end point

        This is pretty messy - first we have to work out the graph associated
        to the first point

        Then we have to iterate over each of the plotters, identify their
        axes, and change the range of the axes to match the screen region
        selected.
        """

        # safety net
        if self.grabpos is None or endpos is None:
            return

        # get points corresponding to corners of rectangle
        pt1 = self.grabpos
        pt2 = endpos

        # work out whether it's worthwhile to zoom: only zoom if there
        # are >=5 pixels movement
        if abs((pt2-pt1).x()) < 10 or abs((pt2-pt1).y()) < 10:
            return

        # scale into graph coordinates
        pt1 /= self.painthelper.cgscale
        pt2 /= self.painthelper.cgscale

        # try to work out in which widget the first point is in
        widget = self.painthelper.pointInWidgetBounds(
            pt1.x(), pt1.y(), widgets.Graph)
        if widget is None:
            return

        # convert points on plotter to points on axis for each axis
        # we also add a neighbouring pixel for the rounding calculation
        xpts = N.array( [pt1.x(), pt2.x(), pt1.x()+1, pt2.x()-1] )
        ypts = N.array( [pt1.y(), pt2.y(), pt1.y()+1, pt2.y()-1] )

        # build up operation list to do zoom
        operations = []

        axes = {}
        # iterate over children, to look for plotters
        for c in [i for i in widget.children if
                  isinstance(i, widgets.GenericPlotter)]:

            # get axes associated with plotter
            caxes = c.parent.getAxes( (c.settings.xAxis,
                                      c.settings.yAxis) )

            for a in caxes:
                if a:
                    axes[a] = True

        # iterate over each axis, and update the ranges
        for axis in axes:
            s = axis.settings
            if s.direction == 'horizontal':
                p = xpts
            else:
                p = ypts

            # convert points on plotter to axis coordinates
            # FIXME: Need To Trap Conversion Errors!
            try:
                r = axis.plotterToGraphCoords(
                    self.painthelper.widgetBounds(axis), p)
            except KeyError:
                continue

            # invert if min and max are inverted
            if r[1] < r[0]:
                r[1], r[0] = r[0], r[1]
                r[3], r[2] = r[2], r[3]

            # build up operations to change axis
            if s.min != r[0]:
                operations.append( document.OperationSettingSet(
                        s.get('min'),
                        utils.round2delt(r[0], r[2])) )

            if s.max != r[1]:
                operations.append( document.OperationSettingSet(
                        s.get('max'),
                        utils.round2delt(r[1], r[3])) )

        # finally change the axes
        self.document.applyOperation(
            document.OperationMultiple(operations,descr=_('zoom axes')) )

    def axesForPoint(self, mousepos):
        """Find all the axes which contain the given mouse position"""

        if self.painthelper is None:
            return []

        pos = self.mapToScene(mousepos)
        px = pos.x() / self.painthelper.cgscale
        py = pos.y() / self.painthelper.cgscale

        axes = []
        for widget, bounds in self.painthelper.widgetBoundsIterator(
            widgettype=widgets.Axis):
            # if widget is axis, and point lies within bounds
            if ( px>=bounds[0] and px<=bounds[2] and
                 py>=bounds[1] and py<=bounds[3] ):

                # convert correct pointer position
                if widget.settings.direction == 'horizontal':
                    val = px
                else:
                    val = py
                coords=widget.plotterToGraphCoords(bounds, N.array([val]))
                axes.append( (widget, coords[0]) )

        return axes

    def emitPicked(self, pickinfo):
        """Report that a new point has been picked"""

        self.pickerinfo = pickinfo
        self.pickeritem.setPos(
            pickinfo.graphpos[0] * self.painthelper.cgscale,
            pickinfo.graphpos[1] * self.painthelper.cgscale)
        self.sigPointPicked.emit(pickinfo)

    def doPick(self, mousepos):
        """Find the point on any plot-like widget closest to the cursor"""

        self.pickerwidgets = []

        pickinfo = widgets.PickInfo()
        # get scalable graph coordinates for mouse point
        pos = self.mapToScene(mousepos)
        pos /= self.painthelper.cgscale

        for w, bounds in self.painthelper.widgetBoundsIterator():
            # ask the widget for its (visually) closest point to the cursor
            try:
                info = w.pickPoint(pos.x(), pos.y(), bounds)
            except AttributeError:
                # widget isn't pickable
                continue

            if info:
                # this is a pickable widget, so remember it for future
                # key navigation
                self.pickerwidgets.append(w)

                if info.distance < pickinfo.distance:
                    # and remember the overall closest
                    pickinfo = info

        if not pickinfo:
            self.pickeritem.hide()
            return

        self.emitPicked(pickinfo)

    def slotBecomeScrollClick(self):
        """If the click is still down when this timer is reached then
        we turn the click into a scrolling click."""

        if self.currentclickmode == 'select':
            qt4.QApplication.setOverrideCursor(qt4.QCursor(qt4.Qt.SizeAllCursor))
            self.currentclickmode = 'scroll'

    def mousePressEvent(self, event):
        """Allow user to drag window around."""

        qt4.QGraphicsView.mousePressEvent(self, event)
        if self.painthelper is None:
            return

        # work out whether user is clicking on a control point
        items = self.items(event.pos())
        self.ignoreclick = ( len(items)==0 or
                             items[0] is not self.pixmapitem or
                             self.painthelper is None )

        if event.button() == qt4.Qt.LeftButton and not self.ignoreclick:

            # need to copy position, otherwise it gets reused!
            self.winpos = qt4.QPoint(event.pos())
            self.grabpos = self.mapToScene(self.winpos)

            if self.clickmode == 'select':
                # we set this to true unless the timer runs out (400ms),
                # then it becomes a scroll click
                # scroll clicks drag the window around, and selecting clicks
                # select widgets!
                self.scrolltimer.start(400)

            elif self.clickmode == 'pick':
                self.pickeritem.show()
                self.pickeritem.setFocus(qt4.Qt.MouseFocusReason)
                self.doPick(event.pos())

            elif self.clickmode == 'scroll':
                qt4.QApplication.setOverrideCursor(
                    qt4.QCursor(qt4.Qt.SizeAllCursor))

            elif self.clickmode == 'graphzoom':
                self.zoomrect.setRect(
                    self.grabpos.x(), self.grabpos.y(),
                    0, 0)
                self.zoomrect.show()

                #self.label.drawRect(self.grabpos, self.grabpos)

            # record what mode we were clicked in
            self.currentclickmode = self.clickmode

    def mouseMoveEvent(self, event):
        """Scroll window by how much the mouse has moved since last time."""

        qt4.QGraphicsView.mouseMoveEvent(self, event)
        if self.painthelper is None:
            return

        if self.currentclickmode == 'scroll':
            event.accept()

            # move scroll bars by amount
            pos = event.pos()
            dx = self.winpos.x()-pos.x()
            scrollx = self.horizontalScrollBar()
            scrollx.setValue( scrollx.value() + dx )

            dy = self.winpos.y()-pos.y()
            scrolly = self.verticalScrollBar()
            scrolly.setValue( scrolly.value() + dy )

            # need to copy point
            self.winpos = qt4.QPoint(event.pos())

        elif self.currentclickmode == 'graphzoom' and self.grabpos is not None:
            pos2 = self.mapToScene(event.pos())
            self.zoomrect.setRect(qt4.QRectF(
                qt4.QPointF(
                    min(self.grabpos.x(), pos2.x()),
                    min(self.grabpos.y(), pos2.y())),
                qt4.QPointF(
                    max(self.grabpos.x(), pos2.x()),
                    max(self.grabpos.y(), pos2.y())),
                ))

        elif self.clickmode == 'select' or self.clickmode == 'pick':
            # find axes which map to this position
            axes = self.axesForPoint(event.pos())
            vals = dict([ (a[0].name, a[1]) for a in axes ])

            self.sigAxisValuesFromMouse.emit(vals)

            if self.currentclickmode == 'pick':
                # drag the picker around
                self.doPick(event.pos())

    def mouseReleaseEvent(self, event):
        """If the mouse button is released, check whether the mouse
        clicked on a widget, and emit a sigWidgetClicked(widget,mode)."""

        qt4.QGraphicsView.mouseReleaseEvent(self, event)
        if self.painthelper is None:
            return

        if event.button() == qt4.Qt.LeftButton and not self.ignoreclick:
            event.accept()
            self.scrolltimer.stop()
            if self.currentclickmode == 'select':
                # work out where the mouse clicked and choose widget
                pos = self.mapToScene(event.pos())
                self.identifyAndClickWidget(pos.x(), pos.y(), event.modifiers())
            elif self.currentclickmode == 'scroll':
                # return the cursor to normal after scrolling
                self.clickmode = 'select'
                self.currentclickmode = None
                qt4.QApplication.restoreOverrideCursor()
            elif self.currentclickmode == 'graphzoom':
                self.zoomrect.hide()
                self.doZoomRect(self.mapToScene(event.pos()))
                self.grabpos = None
            elif self.currentclickmode == 'viewgetclick':
                self.clickmode = 'select'
            elif self.currentclickmode == 'pick':
                self.currentclickmode = None

    def keyPressEvent(self, event):
        """Keypad motion moves the picker if it has focus"""
        if self.pickeritem.hasFocus():

            k = event.key()
            if k == qt4.Qt.Key_Left or k == qt4.Qt.Key_Right:
                # navigate to the previous or next point on the curve
                event.accept()
                dir = 'right' if k == qt4.Qt.Key_Right else 'left'
                ix = self.pickerinfo.index
                pickinfo = self.pickerinfo.widget.pickIndex(
                    ix, dir, self.painthelper.widgetBounds(
                        self.pickerinfo.widget))
                if pickinfo:
                    # more points visible in this direction
                    self.emitPicked(pickinfo)
                return

            elif k == qt4.Qt.Key_Up or k == qt4.Qt.Key_Down:
                # navigate to the next plot up or down on the screen
                event.accept()
                p = self.pickeritem.pos()

                oldw = self.pickerinfo.widget
                pickinfo = widgets.PickInfo()

                dist = float('inf')
                for w in self.pickerwidgets:
                    if w == oldw:
                        continue

                    # ask the widgets to pick their point which is closest horizontally
                    # to the last (screen) x value picked
                    pi = w.pickPoint(
                        self.pickerinfo.graphpos[0], p.y(),
                        self.painthelper.widgetBounds(w),
                        distance='horizontal')
                    if not pi:
                        continue

                    dy = p.y() - pi.graphpos[1]

                    # take the new point which is closest vertically to the current
                    # one and either above or below it as appropriate
                    if abs(dy) < dist and ( (k == qt4.Qt.Key_Up and dy > 0)
                            or (k == qt4.Qt.Key_Down and dy < 0) ):
                        pickinfo = pi
                        dist = abs(dy)

                if pickinfo:
                    oldx = self.pickerinfo.graphpos[0]
                    self.emitPicked(pickinfo)

                    # restore the previous x-position, so that vertical navigation
                    # stays repeatable
                    pickinfo.graphpos = (oldx, pickinfo.graphpos[1])

                return

        # handle up-stream
        qt4.QGraphicsView.keyPressEvent(self, event)

    def wheelEvent(self, event):
        """For zooming in or moving."""

        if event.modifiers() & qt4.Qt.ControlModifier:
            # zoom in/out with ctrl held down
            d = event.angleDelta()
            delta = d.x() if d.x() != 0 else d.y()
            self.sumwheeldelta += delta
            while self.sumwheeldelta <= -120:
                self.slotViewZoomOut()
                self.sumwheeldelta += 120
            while self.sumwheeldelta >= 120:
                self.slotViewZoomIn()
                self.sumwheeldelta -= 120
            event.accept()
        elif event.modifiers() & qt4.Qt.ShiftModifier:
            # scroll horizontally if shift is held down
            d = event.pixelDelta()
            if d.isNull():
                # Fallback mode to angleDelta
                d = event.angleDelta()
            delta = d.x() if d.x() != 0 else d.y()
            scrollx = self.horizontalScrollBar()
            scrollx.setValue(scrollx.value() + delta)
            event.accept()
        else:
            qt4.QGraphicsView.wheelEvent(self, event)

    def identifyAndClickWidget(self, x, y, modifier):
        """Work out which widget was clicked, and if necessary send
        a sigWidgetClicked(widget) signal."""

        if self.document.getNumberPages() == 0:
            return

        widget = self.painthelper.identifyWidgetAtPoint(
            x, y, antialias=self.antialias)
        if widget is None:
            # select page if nothing clicked
            widget = self.document.basewidget.getPage(self.pagenumber)

        # tell connected objects that widget was clicked
        if widget is not None:
            if modifier & qt4.Qt.ControlModifier:
                mode = 'toggle'
            elif modifier & qt4.Qt.ShiftModifier:
                mode = 'add'
            else:
                mode = 'new'

            self.sigWidgetClicked.emit(widget, mode)

    def setPageNumber(self, pageno):
        """Move the the selected page."""

        # we don't need to do anything
        if (self.pagenumber == pageno and
            self.document.changeset == self.docchangeset):
            return

        # keep within bounds
        pageno = min(pageno, self.document.getNumberPages()-1)
        pageno = max(0, pageno)

        self.pagenumber = pageno
        if self.pagenumber != self.oldpagenumber and self.interval != 0:
            self.checkPlotUpdate()

    def getPageNumber(self):
        """Get the the selected page."""
        return self.pagenumber

    @qt4.pyqtSlot(int)
    def slotDocModified(self, ismodified):
        """Update plot on document being modified."""
        # only update if doc is modified and the update policy is set
        # to update on document updates
        if ismodified and self.interval == -1:
            self.checkPlotUpdate()

    def getDevicePixelRatio(self):
        """Get device pixel ratio for window."""
        # ugly, as we have to get it from QWindow
        widget = self
        while widget:
            window = widget.windowHandle()
            if window is not None:
                try:
                    return window.devicePixelRatio()
                except AttributeError:
                    return 1
            widget = widget.parent()
        return 1

    def checkPlotUpdate(self):
        """Check whether plot needs updating."""

        # no threads, so can't get interrupted here
        # draw data into background pixmap if modified

        # is an update required?
        if ( self.zoomfactor == self.oldzoom and
             self.document.changeset == self.docchangeset and
             self.pagenumber == self.oldpagenumber ):
            return

        self.pickeritem.hide()

        # do we need the following line?
        self.pagenumber = min(
            self.document.getNumberPages()-1, self.pagenumber)
        self.oldpagenumber = self.pagenumber

        if self.pagenumber >= 0:
            devicepixelratio = self.getDevicePixelRatio()
            scaling = self.zoomfactor*devicepixelratio
            size = self.document.pageSize(
                self.pagenumber, scaling=scaling, integer=False)

            # draw the data into the buffer
            # errors cause an exception window to pop up
            try:
                phelper = document.PaintHelper(
                    self.document, size,
                    scaling=scaling,
                    dpi=self.dpi,
                    devicepixelratio=devicepixelratio)
                self.document.paintTo(phelper, self.pagenumber)

            except Exception:
                # stop updates this time round and show exception dialog
                d = exceptiondialog.ExceptionDialog(sys.exc_info(), self)
                self.oldzoom = self.zoomfactor
                self.docchangeset = self.document.changeset
                d.exec_()

            self.painthelper = phelper
            self.rendercontrol.addJob(phelper)
        else:
            self.painthelper = None
            self.pagenumber = 0
            size = self.document.docSize()
            pixmap = qt4.QPixmap(*size)
            pixmap.fill( setting.settingdb.color('page') )
            self.setSceneRect(0, 0, *size)
            self.pixmapitem.setPixmap(pixmap)

        self.sigUpdatePage.emit(self.pagenumber)
        self.updatePageToolbar()

        self.updateControlGraphs(self.lastwidgetsselected)
        self.oldzoom = self.zoomfactor
        self.docchangeset = self.document.changeset

    def slotRenderFinished(self, jobid, img, helper):
        """Update image on display if rendering (usually in other
        thread) finished."""
        dpr = helper.devicepixelratio
        bufferpixmap = qt4.QPixmap.fromImage(img)
        bufferpixmap.setDevicePixelRatio(dpr)
        self.setSceneRect(0, 0, bufferpixmap.width()/dpr, bufferpixmap.height()/dpr)
        self.pixmapitem.setPixmap(bufferpixmap)

    def updatePlotSettings(self):
        """Update plot window settings from settings."""
        self.setTimeout(setting.settingdb['plot_updatepolicy'])
        self.antialias = setting.settingdb['plot_antialias']
        self.rendercontrol.updateNumberThreads()
        self.actionForceUpdate()

    def contextMenuEvent(self, event):
        """Show context menu."""

        menu = qt4.QMenu(self)

        # add some useful entries
        menu.addAction( self.vzactions['view.zoommenu'] )
        menu.addSeparator()
        menu.addAction( self.vzactions['view.prevpage'] )
        menu.addAction( self.vzactions['view.nextpage'] )
        menu.addSeparator()

        # force an update now menu item
        menu.addAction(_('Force update'), self.actionForceUpdate)

        if self.isfullscreen:
            menu.addAction(_('Close full screen'), self.slotFullScreen)
        else:
            menu.addAction( self.vzactions['view.fullscreen'] )

        # Update policy submenu
        submenu = menu.addMenu(_('Updates'))
        intgrp = qt4.QActionGroup(self)

        # bind interval options to actions
        for intv, text in self.updateintervals:
            act = intgrp.addAction(text)
            act.setCheckable(True)
            def setfn(interval):
                return lambda checked: self.actionSetTimeout(interval, checked)
            act.triggered.connect(setfn(intv))
            if intv == self.interval:
                act.setChecked(True)
            submenu.addAction(act)

        # antialias
        menu.addSeparator()
        act = menu.addAction(_('Antialias'), self.actionAntialias)
        act.setCheckable(True)
        act.setChecked(self.antialias)

        menu.exec_(qt4.QCursor.pos())

    def actionForceUpdate(self):
        """Force an update for the graph."""
        self.docchangeset = -100
        self.checkPlotUpdate()

    def slotFullScreen(self):
        """Show window full screen or not."""
        if not self.isfullscreen:
            self._fullscreenwindow = FullScreenPlotWindow(
                self.document, self.pagenumber)
        else:
            # cheesy way of closing full screen window
            p = self
            while p.parent() is not None:
                p = p.parent()
            p.close()

    def setTimeout(self, interval):
        """Change timer setting without changing save value."""
        self.interval = interval
        if interval <= 0:
            # stop updates
            if self.timer.isActive():
                self.timer.stop()
        else:
            # change interval to one selected
            self.timer.setInterval(interval)
            # start timer if it was stopped
            if not self.timer.isActive():
                self.timer.start()

    def actionSetTimeout(self, interval, checked):
        """Called by setting the interval."""

        self.setTimeout(interval)

        # remember changes for next time
        setting.settingdb['plot_updatepolicy'] = self.interval

    def actionAntialias(self):
        """Toggle antialias."""
        self.antialias = not self.antialias
        setting.settingdb['plot_antialias'] = self.antialias
        self.actionForceUpdate()

    def setZoomFactor(self, zoomfactor):
        """Set the zoom factor of the window."""
        zoomfactor = max(0.05, min(20, zoomfactor))
        self.zoomfactor = float(zoomfactor)
        self.checkPlotUpdate()

    def slotViewZoomIn(self):
        """Zoom into the plot."""
        self.setZoomFactor(self.zoomfactor * N.sqrt(2.))

    def slotViewZoomOut(self):
        """Zoom out of the plot."""
        self.setZoomFactor(self.zoomfactor / N.sqrt(2.))

    def slotViewZoomWidth(self):
        """Make the zoom factor so that the plot fills the whole width."""

        # need to take account of scroll bars when deciding size
        viewportsize = self.maximumViewportSize()
        aspectwin = viewportsize.width() / viewportsize.height()
        r = self.pixmapitem.boundingRect()
        aspectplot = r.width() / r.height()

        width = viewportsize.width()
        if aspectwin > aspectplot:
            # take account of scroll bar
            width -= self.verticalScrollBar().width()

        mult = width / r.width()
        self.setZoomFactor(self.zoomfactor * mult)

    def slotViewZoomHeight(self):
        """Make the zoom factor so that the plot fills the whole width."""

        viewportsize = self.maximumViewportSize()
        pixrect = self.pixmapitem.boundingRect()

        try:
            aspectwin = viewportsize.width() / viewportsize.height()
            aspectplot = pixrect.width() / pixrect.height()
        except ZeroDivisionError:
            return

        height = viewportsize.height()
        if aspectwin < aspectplot:
            # take account of scroll bar
            height -= self.horizontalScrollBar().height()

        mult = height / pixrect.height()
        self.setZoomFactor(self.zoomfactor * mult)

    def slotViewZoomPage(self):
        """Make the zoom factor correct to show the whole page."""

        viewportsize = self.maximumViewportSize()
        r = self.pixmapitem.boundingRect()
        if r.width() != 0 and r.height() != 0:
            multw = viewportsize.width() / r.width()
            multh = viewportsize.height() / r.height()
            self.setZoomFactor(self.zoomfactor * min(multw, multh))

    def slotViewZoom11(self):
        """Restore the zoom to 1:1"""
        self.setZoomFactor(1.)

    def slotViewPreviousPage(self):
        """View the previous page."""
        self.setPageNumber( self.pagenumber - 1 )

    def slotViewNextPage(self):
        """View the next page."""
        self.setPageNumber( self.pagenumber + 1 )

    def updatePageToolbar(self):
        """Update page number when the plot window says so."""

        # disable previous and next page actions
        if self.vzactions is not None:
            np = self.document.getNumberPages()
            self.vzactions['view.prevpage'].setEnabled(self.pagenumber != 0)
            self.vzactions['view.nextpage'].setEnabled(self.pagenumber < np-1)

    def slotSelectMode(self, action):
        """Called when the selection mode has changed."""

        modecnvt = { self.vzactions['view.select'] : 'select',
                     self.vzactions['view.pick'] : 'pick',
                     self.vzactions['view.zoomgraph'] : 'graphzoom' }

        # close the current picker
        self.pickeritem.hide()
        self.sigPickerEnabled.emit(False)

        # convert action into clicking mode
        self.clickmode = modecnvt[action]

        if self.clickmode == 'select':
            self.pixmapitem.unsetCursor()
        elif self.clickmode == 'graphzoom':
            self.pixmapitem.setCursor(qt4.Qt.CrossCursor)
        elif self.clickmode == 'pick':
            self.pixmapitem.setCursor(qt4.Qt.CrossCursor)
            self.sigPickerEnabled.emit(True)

    def getClick(self):
        """Return a click point from the graph."""

        # wait for click from user
        qt4.QApplication.setOverrideCursor(qt4.QCursor(qt4.Qt.CrossCursor))
        oldmode = self.clickmode
        self.clickmode = 'viewgetclick'
        while self.clickmode == 'viewgetclick':
            qt4.qApp.processEvents()
        self.clickmode = oldmode
        qt4.QApplication.restoreOverrideCursor()

        # take clicked point and convert to coords of scrollview
        pt = self.grabpos

        # try to work out in which widget the first point is in
        widget = self.painthelper.pointInWidgetBounds(
            pt.x(), pt.y(), widgets.Graph)
        if widget is None:
            return []

        # convert points on plotter to points on axis for each axis
        xpts = N.array( [pt.x()] )
        ypts = N.array( [pt.y()] )

        axesretn = []
        # iterate over children, to look for plotters
        for c in [i for i in widget.children if
                  isinstance(i, widgets.GenericPlotter)]:

            # get axes associated with plotter
            axes = c.parent.getAxes( (c.settings.xAxis,
                                      c.settings.yAxis) )

            # iterate over each, and update the ranges
            for axis in [a for a in axes if a is not None]:
                s = axis.settings
                if s.direction == 'horizontal':
                    p = xpts
                else:
                    p = ypts

                # convert point on plotter to axis coordinate
                # FIXME: Need To Trap Conversion Errors!
                r = axis.plotterToGraphCoords(
                    self.painthelper.widgetBounds(axis), p)

                axesretn.append( (axis.path, r[0]) )

        return axesretn

    def selectedWidgets(self, widgets):
        """Update control items on screen associated with widget.
        Called when widgets have been selected in the tree edit window
        """
        self.updateControlGraphs(widgets)
        self.lastwidgetsselected = widgets

    def updateControlGraphs(self, widgets):
        """Add control graphs for the widgets given."""

        # delete old items from root
        for c in list(self.controlgraphroot.childItems()):
            self.scene.removeItem(c)

        # add each item to the root
        if self.painthelper:
            for widget in widgets:
                cgis = self.painthelper.getControlGraph(widget)
                if cgis:
                    for control in cgis:
                        control.createGraphicsItem(self.controlgraphroot)

class FullScreenPlotWindow(qt4.QScrollArea):
    """Window for showing plot in full-screen mode."""

    def __init__(self, document, pagenumber):
        qt4.QScrollArea.__init__(self)
        self.setFrameShape(qt4.QFrame.NoFrame)
        self.setWidgetResizable(True)

        # window which shows plot
        self.document = document
        pw = self.plotwin = PlotWindow(document, None)
        pw.isfullscreen = True
        pw.pagenumber = pagenumber
        self.setWidget(pw)
        pw.setFocus()

        self.showFullScreen()

        self.toolbar = qt4.QToolBar(_("Full screen toolbar"), self)
        self.toolbar.addAction(utils.getIcon("kde-window-close"), _("Close"),
                               self.close)
        for a in ('view.zoom11', 'view.zoomin', 'view.zoomout',
                  'view.zoomwidth', 'view.zoomheight',
                  'view.zoompage', 'view.prevpage', 'view.nextpage'):
            self.toolbar.addAction( pw.vzactions[a] )
        self.toolbar.show()

    def resizeEvent(self, event):
        """Make zoom fit screen."""

        qt4.QScrollArea.resizeEvent(self, event)

        # size graph to fill screen
        pagesize = self.document.pageSize(self.plotwin.pagenumber,
                                          dpi=self.plotwin.dpi)
        screensize = self.plotwin.size()

        aspectw = screensize.width() / pagesize[0]
        aspecth = screensize.height() / pagesize[1]

        self.plotwin.zoomfactor = min(aspectw, aspecth)
        self.plotwin.checkPlotUpdate()

    def keyPressEvent(self, event):

        k = event.key()
        if k == qt4.Qt.Key_Escape:
            event.accept()
            self.close()
            return
        qt4.QScrollArea.keyPressEvent(self, event)



#    Copyright (C) 2016 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

from __future__ import division, print_function
import re

from .. import qtall as qt4
from ..compat import cstr
from .. import utils
from ..dialogs import importdialog
from . import defn_nd
from . import simpleread
from . import dialog_csv

def _(text, disambiguation=None, context="Import_ND"):
    return qt4.QCoreApplication.translate(context, text, disambiguation)

class ImportTabND(importdialog.ImportTab):
    """Tab for importing from a ND data file."""

    resource = "import_nd.ui"

    def loadUi(self):
        """Load user interface and set up validators."""
        importdialog.ImportTab.loadUi(self)

        self.nd_mode.defaultlist = [_("Text"), _("CSV")]
        self.nd_shapeedit.default = [_("Auto")]
        self.nd_csvdelim.default = dialog_csv.csv_delimiters
        self.nd_csvtextdelim.default = dialog_csv.csv_text_delimiters
        self.nd_csvlocale.defaultlist = dialog_csv.csv_locales

        self.nd_mode.currentIndexChanged.connect(self.slotNewMode)

    def slotNewMode(self, index):
        """Change other widgets depending on mode."""
        csv = index == 1
        self.nd_csvdelim.setEnabled(csv)
        self.nd_csvtextdelim.setEnabled(csv)
        self.nd_csvlocale.setEnabled(csv)

    def reset(self):
        """Reset controls."""
        self.nd_datasetedit.setEditText("")
        self.nd_shapeedit.setEditText(_("Auto"))
        self.nd_transposecheck.setChecked(False)
        self.nd_mode.setCurrentIndex(0)
        self.nd_csvdelim.setEditText(dialog_csv.csv_delimiters[0])
        self.nd_csvtextdelim.setEditText(dialog_csv.csv_text_delimiters[0])
        self.nd_csvlocale.setCurrentIndex(0)

    def doPreview(self, filename, encoding):
        """Preview nD dataset files."""

        try:
            ifile = utils.openEncoding(filename, encoding)
            text = ifile.read(4096) + "\n"
            if len(ifile.read(1)) != 0:
                # if there is remaining data add ...
                text += "...\n"
            self.nd_previewedit.setPlainText(text)
            return True

        except (UnicodeError, EnvironmentError):
            self.nd_previewedit.setPlainText("")
            return False

    def doImport(self, doc, filename, linked, encoding, prefix, suffix, tags):
        """Import from ND file."""

        class error(RuntimeError):
            pass

        try:
            dataset = self.nd_datasetedit.text().strip()
            if not dataset:
                raise error(_("A dataset name should be given"))

            transpose = self.nd_transposecheck.isChecked()

            mode = ("text", "csv")[self.nd_mode.currentIndex()]
            # this needs to be a str for the csv module (py2)
            csvdelimiter = str(self.nd_csvdelim.text())
            if csvdelimiter in dialog_csv.csv_delimiter_map:
                csvdelimiter = dialog_csv.csv_delimiter_map[csvdelimiter]
            csvtextdelimiter = str(self.nd_csvtextdelim.text())
            csvlocale = dialog_csv.csvLocaleIndexToLocale(
                self.nd_csvlocale.currentIndex())

            shapetxt = self.nd_shapeedit.text().strip()
            if shapetxt == _("Auto"):
                shape = None
            else:
                shapesplit = re.split("[,;x* ]+", shapetxt)
                try:
                    shape = tuple([int(x) for x in shapesplit])
                except ValueError:
                    raise error(_("Shape entries should be integers"))
                if len(shape) == 0:
                    raise error(_("No shape entries given"))

            params = defn_nd.ImportParamsND(
                dataset=dataset,
                filename=filename,
                shape=shape,
                transpose=transpose,
                mode=mode,
                csvdelimiter=csvdelimiter,
                csvtextdelimiter=csvtextdelimiter,
                csvlocale=csvlocale,
                prefix=prefix, suffix=suffix,
                tags=tags,
                linked=linked,
                encoding=encoding
                )

            # do the importing
            op = defn_nd.OperationDataImportND(params)
            doc.applyOperation(op)

            # show result
            output = [_("Successfully read:")]
            for ds in op.outnames:
                output.append("%s: %s" % (
                    ds,
                    doc.data[ds].description())
                )
            output = "\n".join(output)

            # feature feedback
            utils.feedback.importcts['nd'] += 1

        except error as e:
            output = e.args[0]

        except simpleread.ReadNDError as e:
            output = _("Error importing datasets:\n %s") % cstr(e)

        # show status in preview box
        self.nd_previewedit.setPlainText(output)

importdialog.registerImportTab(_("&ND"), ImportTabND)



# commandinterface.py
# this module supplies the command line interface for plotting

#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""
Module supplies the command interface used in the program, and for
external programs.
"""

from __future__ import division, print_function
import os.path
import numpy as N

from ..compat import cbasestr
from .. import qtall as qt4
from .. import setting
from .. import embed
from .. import plugins
from .. import utils
from .. import datasets

from . import operations
from . import mime
from . import export

def _(text, disambiguation=None, context='CommandInterface'):
    """Translate text."""
    return qt4.QCoreApplication.translate(context, text, disambiguation)

def registerImportCommand(name, method, filenamearg=0):
    """Add command to command interface."""
    setattr(CommandInterface, name, method)
    CommandInterface.import_commands.append(name)
    CommandInterface.import_filenamearg[name] = filenamearg

class CommandInterface(qt4.QObject):
    """Class provides command interface."""

    # commands which are safe in any script (excluding import commands)
    safe_commands = [
        'Action',
        'Add',
        'AddCustom',
        'AddImportPath',
        'CloneWidget',
        'CreateHistogram',
        'DatasetPlugin',
        'FilterDatasets',
        'Get',
        'GetChildren',
        'GetColormap',
        'GetData',
        'GetDataType',
        'GetDatasets',
        'ImportFITSFile',
        'List',
        'NodeChildren',
        'NodeType',
        'ReloadData',
        'Remove',
        'RemoveCustom',
        'Rename',
        'ResolveReference',
        'Set',
        'SetData',
        'SetData2D',
        'SetData2DExpression',
        'SetData2DExpressionXYZ',
        'SetData2DXYFunc',
        'SetDataDateTime',
        'SetDataExpression',
        'SetDataND',
        'SetDataRange',
        'SetDataText',
        'SetToReference',
        'SetVerbose',
        'SettingType',
        'TagDatasets',
        'To',
        'WidgetType',
        ]

    # commands for importing data
    import_commands = []
    # number of argument which contains filename
    import_filenamearg = {}

    # commands which can modify disk, etc
    unsafe_commands = [
        'Export',
        'Print',
        'Save',
        ]

    def __init__(self, document):
        """Initialise the interface."""
        qt4.QObject.__init__(self)

        self.document = document
        self.currentwidget = self.document.basewidget
        self.verbose = False
        self.importpath = []

        self.document.sigWiped.connect(self.slotWipedDoc)

        self.Root = embed.WidgetNode(self, 'widget', '/')

    @qt4.pyqtSlot()
    def slotWipedDoc(self):
        """When the document is wiped, we change to the root widget."""
        self.To('/')

    def findFileOnImportPath(self, filename):
        """Find file on path, returning filename, or original if not found."""
        for path in self.importpath:
            fname = os.path.join(path, filename)
            try:
                # try to open file to make sure we have access to it and
                # it exists
                opened = open(fname)
                opened.close()
                return fname
            except EnvironmentError:
                pass
        return filename

    def SetVerbose(self, v=True):
        """Specify whether we want verbose output after operations."""
        self.verbose = v

    def Add(self, widgettype, **args_opt):
        """Add a widget to the widget with the type given.

        optional arguments:
          widget: widget path to place widget in
          autoadd: if True (default), any subwidgets associated with
            widget are created automatically
          setting names, e.g. leftMargin='1cm', Border_color='red'
        """

        at = self.currentwidget
        if 'widget' in args_opt:
            at = self.document.resolveWidgetPath(
                self.currentwidget, args_opt['widget'])
            del args_opt['widget']

        op = operations.OperationWidgetAdd(at, widgettype, **args_opt)
        w = self.document.applyOperation(op)

        if self.verbose:
            print(_("Added a widget of type '%s' (%s)") % (type, w.userdescription))

        return w.name

    def AddCustom(self, ctype, name, val, mode='appendalways'):
        """Add a custom definition for evaluation of expressions.
	This can define a constant (can be in terms of other
	constants), a function of 1 or more variables, or a function
	imported from an external python module.

        ctype is "constant", "function", "definition" (either
        constant or function), "import", "color" or "colormap".

        name is name of constant, color or colormap, "function(x, y,
        ...)"  or module name.

        val is definition for constant or function (both are
        _strings_), or is a list of symbols for a module (comma
        separated items in a string). For a colormap, val is a list of
        4-item tuples containing R,G,B,alpha values from 0 to 255. For
        a color this is a string with the format '#RRGGBB' or
        '#RRGGBBAA'.

        if mode is 'appendalways', the custom value is appended to the
        end of the list even if there is one with the same name. If
        mode is 'replace', it replaces any existing definition in the
        same place in the list or is appended otherwise. If mode is
        'append', then an existing definition is deleted, and the new
        one appended to the end.

        """

        if ctype == 'colormap':
            self.document.evaluate.validateProcessColormap(val)
        else:
            if not isinstance(val, cbasestr):
                raise RuntimeError('Value should be string')

        if mode not in ('appendalways', 'append', 'replace'):
            raise RuntimeError('Invalid mode')

        try:
            attr = operations.OperationSetCustom.type_to_attr[ctype]
        except KeyError:
            raise RuntimeError('Invalid type')

        vals = list(getattr(self.document.evaluate, attr))

        item = [name.strip(), val]
        if mode == 'appendalways':
            vals.append(item)
        else:
            # find any existing item
            for i, (n, v) in enumerate(vals):
                if n == name:
                    if mode == 'append':
                        del vals[i]
                        vals.append(item)
                    else: # replace
                        vals[i] = item
                    break
            else:
                # no existing item, so append
                vals.append(item)

        op = operations.OperationSetCustom(ctype, vals)
        self.document.applyOperation(op)

    def AddImportPath(self, directory):
        """Add directory to import file path."""
        assert isinstance(directory, cbasestr)
        self.importpath.append(directory)

    def CloneWidget(self, widget, newparent, newname=None):
        """Clone the widget given, placing the copy in newparent and
        the name given.

        newname is an optional new name to give it

        Returns new widget path
        """

        widget = self.document.resolveWidgetPath(self.currentwidget, widget)
        newparent = self.document.resolveWidgetPath(self.currentwidget, newparent)
        op = mime.OperationWidgetClone(widget, newparent, newname)
        w = self.document.applyOperation(op)
        return w.path

    def CreateHistogram(self, inexpr, outbinsds, outvalsds, binparams=None,
                        binmanual=None, method='counts',
                        cumulative = 'none', errors=False):
        """Histogram an input expression.

        inexpr is input expression
        outbinds is the name of the dataset to create giving bin positions
        outvalsds is name of dataset for bin values
        binparams is None or (numbins, minval, maxval, islogbins)
        binmanual is None or a list of bin values
        method is 'counts', 'density', or 'fractions'
        cumulative is to calculate cumulative distributions which is
          'none', 'smalltolarge' or 'largetosmall'
        errors is to calculate Poisson error bars
        """
        op = operations.OperationDatasetHistogram(
            inexpr, outbinsds, outvalsds, binparams=binparams,
            binmanual=binmanual, method=method,
            cumulative=cumulative, errors=errors)
        self.document.applyOperation(op)

        if self.verbose:
            print(_('Constructed histogram of "%s", creating datasets'
                    ' "%s" and "%s"') % (inexpr, outbinsds, outvalsds))

    def DatasetPlugin(self, pluginname, fields, datasetnames={}):
        """Use a dataset plugin.

        pluginname: name of plugin to use
        fields: dict of input values to plugin
        datasetnames: dict mapping old names to new names of datasets
        if they are renamed. The new name None means dataset is deleted."""

        # lookup plugin (urgh)
        plugin = None
        for pkls in plugins.datasetpluginregistry:
            if pkls.name == pluginname:
                plugin = pkls()
                break
        if plugin is None:
            raise RuntimeError("Cannot find dataset plugin '%s'" % pluginname)

        # do the work
        op = operations.OperationDatasetPlugin(plugin, fields,
                                               datasetnames=datasetnames)
        outdatasets = self.document.applyOperation(op)

        if self.verbose:
            print(_("Used dataset plugin %s to make datasets %s") % (
                pluginname, ', '.join(outdatasets)))

    def Remove(self, name):
        """Remove a widget from the dataset."""
        w = self.document.resolveWidgetPath(self.currentwidget, name)
        op = operations.OperationWidgetDelete(w)
        self.document.applyOperation(op)
        if self.verbose:
            print(_("Removed widget '%s'") % name)

    def RemoveCustom(self, name):
        """Removes a custom-defined constant, function or import."""

        # look for definiton and delete if found
        for ctype, attr in (
                ('import', 'def_imports'),
                ('definition', 'def_definitions'),
                ('color', 'def_colors'),
                ('colormap', 'def_colormaps')):
            vals = list(getattr(self.document.evaluate, attr))
            for i, (cname, cval) in enumerate(vals):
                if name == cname:
                    del vals[i]
                    op = operations.OperationSetCustom(ctype, vals)
                    self.document.applyOperation(op)
                    return
        else:
            raise ValueError('Custom variable not defined')

    def To(self, where):
        """Change to a widget within the current widget.

        where is a path to the widget relative to the current widget
        """

        self.currentwidget = self.document.resolveWidgetPath(
            self.currentwidget,
            where)

        if self.verbose:
            print(_("Changed to widget '%s'") % self.currentwidget.path)

    def List(self, where='.'):
        """List the contents of a widget, by default the current widget."""

        widget = self.document.resolveWidgetPath(self.currentwidget, where)
        children = widget.childnames

        if len(children) == 0:
            print('%30s' % _('No children found'))
        else:
            # output format name, type
            for name in children:
                w = widget.getChild(name)
                print('%10s %10s %30s' % (name, w.typename, w.userdescription))

    def Get(self, var):
        """Get the value of a setting."""
        return self.document.resolveSettingPath(self.currentwidget, var).val

    def GetChildren(self, where='.'):
        """Return a list of widgets which are children of the widget of the
        path given."""
        return list(
            self.document.resolveWidgetPath(self.currentwidget, where).childnames )

    def GetColormap(self, name, invert=False, nvals=256):
        """Return an array of [red,green,blue,alpha] values
        representing the colormap with the name given.

        Each return value is between 0 and 255.

        The number of values to return is given by nvals
        """
        cmap = self.document.evaluate.getColormap(name, invert)
        return utils.getColormapArray(cmap, nvals)

    def GetDatasets(self):
        """Return a list of names of datasets."""
        return sorted(self.document.data)

    def ResolveReference(self, setn):
        """If the setting is set to a reference, follow the chain of
        references to return the absolute path to the real setting.

        If it is not a reference return None.
        """

        setn = self.document.resolveSettingPath(self.currentwidget, setn)
        if setn.isReference():
            real = setn.getReference().resolve(setn)
            return real.path
        else:
            return None

    def Save(self, filename, mode='vsz'):
        """Save the state to a file.

        mode can be:
         'vsz': standard veusz text format
         'hdf5': HDF5 format
        """
        self.document.save(filename, mode)

    def Set(self, setting_path, val):
        """Set the value of a setting."""
        setn = self.document.resolveSettingPath(self.currentwidget, setting_path)
        op = operations.OperationSettingSet(setn, val)
        self.document.applyOperation(op)

        if self.verbose:
            print( _("Set setting '%s' to %s") % (setting_path, repr(setn.get())) )

    def SetToReference(self, setting_path, val):
        """Set setting to a reference value."""
        setn = self.document.resolveSettingPath(self.currentwidget, setting_path)
        op = operations.OperationSettingSet(setn, setting.Reference(val))
        self.document.applyOperation(op)

        if self.verbose:
            print( _( "Set setting '%s' to %s") % (setting_path, repr(setn.get())) )

    def SetData(self, name, val, symerr=None, negerr=None, poserr=None):
        """Create/set dataset name with values (and optionally errors)."""

        data = datasets.Dataset(val, symerr, negerr, poserr)
        op = operations.OperationDatasetSet(name, data)
        self.document.applyOperation(op)

        if self.verbose:
            print(
                _("Set dataset '%s':\n"
                  " Values = %s\n"
                  " Symmetric errors = %s\n"
                  " Negative errors = %s\n"
                  " Positive errors = %s") % (
                      name, str(data.data), str(data.serr),
                      str(data.nerr), str(data.perr))
            )

    def SetDataDateTime(self, name, vals):
        """Set datetime dataset to be values given.
        vals is a list of python datetime objects
        """
        v = [utils.datetimeToFloat(x) for x in vals]
        ds = datasets.DatasetDateTime(v)
        op = operations.OperationDatasetSet(name, ds)
        self.document.applyOperation(op)

        if self.verbose:
            print(
                _("Set dataset '%s':\n"
                  " Values = %s") % (
                      name, str(ds.data))
            )

    def SetDataExpression(self, name, val, symerr=None, negerr=None, poserr=None,
                          linked=False, parametric=None):
        """Create a dataset based on text expressions.

        Expressions are functions of existing datasets.
        If evaluating the expression 'y*10' in negerr, then the negerrs of dataset y
        are used, and so on.
        To access a specific part of the dataset y, the suffixes _data, _serr, _perr,
        and _nerr can be appended.

        If linked is True then the expressions are reevaluated if the document
        is modified

        parametric: tuple of (minval, maxval, numitems) for creating parametric
                    datasets. t set to this range when evaluating expressions.

        """

        expr = {'data': val, 'serr': symerr, 'nerr': negerr, 'perr': poserr}
        op = operations.OperationDatasetCreateExpression(name, expr, linked,
                                                         parametric=parametric)

        data = self.document.applyOperation(op)

        if self.verbose:
            print(
                _("Set dataset '%s' based on expression:\n"
                  " Values = %s\n"
                  " Symmetric errors = %s\n"
                  " Negative errors = %s\n"
                  " Positive errors = %s") % (
                      name, str(data.data), str(data.serr),
                      str(data.nerr), str(data.perr))
            )
            if parametric:
                print(_(" Where t goes form %g:%g in %i steps") % parametric)
            print(_(" linked to expression = %s") % repr(linked))

    def SetDataND(self, name, val):
        """Set n-dimensional dataset name with values."""

        data = datasets.DatasetND(val)
        op = operations.OperationDatasetSet(name, data)
        self.document.applyOperation(op)

        if self.verbose:
            print(
                _("Set dataset (nD) '%s':\n"
                  " Values = %s\n") % (
                      name, str(data.data))
            )

    def SetDataRange(self, name, numsteps, val, symerr=None, negerr=None,
                     poserr=None, linked=False):
        """Create dataset based on ranges of values, e.g. 1 to 10 in 10 steps

        name: name of dataset
        numsteps: number of steps to create
        val: range in form of tuple (minval, maxval)
        symerr, negerr & poserr: ranges for errors (optional)
        """

        parts = {'data': val, 'serr': symerr, 'nerr': negerr, 'perr': poserr}
        op = operations.OperationDatasetCreateRange(name, numsteps, parts,
                                                    linked)
        self.document.applyOperation(op)

        if self.verbose:
            print(
                _("Set dataset '%s' based on range:\n"
                  " Number of steps = %i\n"
                  " Range of data = %s\n"
                  " Range of symmetric error = %s\n"
                  " Range of positive error = %s\n"
                  " Range of negative error = %s") % (
                      name, numsteps, repr(val),
                      repr(symerr), repr(poserr), repr(negerr))
              )

    def SetData2DExpression(self, name, expr, linked=False):
        """Create a 2D dataset based on expressions

        name is the new dataset name
        expr is an expression which should return a 2D array
        linked specifies whether to permanently link the dataset to the expressions
        """

        op = operations.OperationDataset2DCreateExpression(name, expr, linked)
        data = self.document.applyOperation(op)

        if self.verbose:
            print(
                _("Set 2D dataset '%s' based on expressions\n"
                  " expression = %s\n"
                  " linked to expression = %s\n"
                  " Made a dataset (%i x %i)") % (
                      name, repr(expr), repr(linked),
                      data.data.shape[0], data.data.shape[1])
            )

    def SetData2DExpressionXYZ(self, name, xexpr, yexpr, zexpr, linked=False):
        """Create a 2D dataset based on expressions in x, y and z

        xexpr is an expression which expands to an equally-spaced grid of x coordinates
        yexpr expands to equally spaced y coordinates
        zexpr expands to z coordinates.
        linked specifies whether to permanently link the dataset to the expressions
        """

        op = operations.OperationDataset2DCreateExpressionXYZ(name, xexpr, yexpr, zexpr,
                                                              linked)
        data = self.document.applyOperation(op)

        if self.verbose:
            print(
                _("Made 2D dataset '%s' based on expressions:\n"
                  " X expression = %s\n"
                  " Y expression = %s\n"
                  " Z expression = %s\n"
                  " is linked to expression = %s\n"
                  " Shape (%i x %i)") % (
                      name,
                      repr(xexpr), repr(yexpr), repr(zexpr),
                      repr(linked),
                      data.data.shape[0], data.data.shape[1])
            )

    def SetData2DXYFunc(self, name, xstep, ystep, expr, linked=False):
        """Create a 2D dataset based on expressions of a range of x and y

        xstep is a tuple(min, max, step)
        ystep is a tuple(min, max, step)
        expr is an expression of x and y
        linked specifies whether to permanently link the dataset to the expressions
        """

        op = operations.OperationDataset2DXYFunc(name, xstep, ystep,
                                                 expr, linked)
        data = self.document.applyOperation(op)

        if self.verbose:
            print(
                _("Set 2D dataset '%s' based on function of x and y\n"
                  " X steps = %s\n"
                  " Y steps = %s\n"
                  " Expression = %s\n"
                  " linked to expression = %s\n"
                  " Made a dataset (%i x %i)") % (
                      name, repr(xstep), repr(ystep),
                      repr(expr), repr(linked),
                      data.data.shape[0], data.data.shape[1])
            )

    def SetData2D(self, name, data, xrange=None, yrange=None,
                  xedge=None, yedge=None,
                  xcent=None, ycent=None):
        """Create a 2D dataset.

        name: name of dataset
        data: 2d array
        xrange: optional tuple with X range of data (min, max)
        yrange: optional tuple with Y range of data (min, max)
        xedge: x values for grid (instead of rangex)
        yedge: y values for grid (instead of rangey)
        xcent: x values for pixel centres (instead of rangex)
        ycent: y values for pixel centres (instead of rangey)
        """

        data = N.array(data)

        if ( (xedge is not None and not utils.checkAscending(xedge)) or
             (yedge is not None and not utils.checkAscending(yedge)) ):
            raise ValueError("xedge and yedge must be ascending, if given")
        if ( (xcent is not None and not utils.checkAscending(xcent)) or
             (ycent is not None and not utils.checkAscending(ycent)) ):
            raise ValueError("xcent and ycent must be ascending, if given")

        if ( (xedge is not None and len(xedge) != data.shape[1]+1) or
             (yedge is not None and len(yedge) != data.shape[0]+1) ):
            raise ValueError("xedge and yedge lengths must be data shape+1")
        if ( (xcent is not None and len(xcent) != data.shape[1]) or
             (ycent is not None and len(ycent) != data.shape[0]) ):
            raise ValueError("xcent and ycent lengths must be data shape")

        data = datasets.Dataset2D(data, xrange=xrange, yrange=yrange,
                                  xedge=xedge, yedge=yedge,
                                  xcent=xcent, ycent=ycent)
        op = operations.OperationDatasetSet(name, data)
        self.document.applyOperation(op)

        if self.verbose:
            print(_("Set 2d dataset '%s'") % name)

    def SetDataText(self, name, val):
        """Create a text dataset."""

        data = datasets.DatasetText(val)
        op = operations.OperationDatasetSet(name, data)
        self.document.applyOperation(op)

        if self.verbose:
            print(
                _("Set text dataset '%s'\n"
                  "Values = %s") % (
                      name, repr(data.data))
            )

    def GetData(self, name):
        """Return the data with the name.

        For a 1D dataset, returns a tuple (None if not defined)
            (data, serr, nerr, perr)
        For a 2D dataset, returns
            (data, xrange, yrange)
        For an nD dataset returns data array
        For a text dataset, return a list of text
        For a date dataset, return a list of python datetime objects

        Return copies, so that the original data can't be indirectly modified
        """

        d = self.document.getData(name)
        if d.displaytype == 'text':
            return d.data[:]
        elif d.displaytype == 'date':
            return [utils.floatToDateTime(x) for x in d.data]
        elif d.dimensions == 2:
            return (d.data.copy(), d.xrange, d.yrange)
        elif d.dimensions == -1:
            return d.data.copy()
        else:
            data = serr = nerr = perr = None
            if d.data is not None:
                data = d.data.copy()
            if d.serr is not None:
                serr = d.serr.copy()
            if d.nerr is not None:
                nerr = d.nerr.copy()
            if d.perr is not None:
                perr = d.perr.copy()

            return (data, serr, nerr, perr)

    def GetDataType(self, name):
        """Return the type of dataset. Returns None if no dataset.

        For a 1D dataset, returns '1d'
        For a 2D dataset, returns '2d'
        For a text dataset, returns 'text'
        For a datetime dataset, returns 'datetime'
        """

        try:
            d = self.document.getData(name)
        except KeyError:
            return None
        if d.displaytype == 'text':
            return 'text'
        elif d.displaytype == 'date':
            return 'datetime'
        elif d.dimensions == 2:
            return '2d'
        else:
            return '1d'

    def ReloadData(self):
        """Reload any linked datasets.

        Returned is a tuple (datasets, errors)
         where datasets is a list of datasets read
         errors is a dict of the datasets with the number of errors while
         converting the data
        """

        return self.document.reloadLinkedDatasets()

    def Action(self, action, widget='.'):
        """Performs action on current widget."""

        w = self.document.resolveWidgetPath(self.currentwidget, widget)

        # run action
        w.getAction(action).function()

    def Print(self):
        """Print document."""
        export.printDialog(None, self.document)

    def Export(self, filename, color=True, page=[0], dpi=100,
               antialias=True, quality=85, backcolor='#ffffff00',
               pdfdpi=150, svgtextastext=False):
        """Export plot to filename.

        color is True or False if color is requested in output file
        page is a list of page numbers to export
        dpi is the number of dots per inch for bitmap output files
        antialias antialiases output if True
        quality is a quality parameter for jpeg output
        backcolor is the background color for bitmap files, which is a name or
         a #RRGGBBAA value (red, green, blue, alpha)
        pdfdpi is the dpi to use when exporting eps or pdf files
        svgtextastext: write text in SVG as text, rather than curves
        """

        # compatibility where page was a single number
        try:
            pages = [p for p in page]
        except TypeError:
            pages = [page]

        e = export.Export(
            self.document, filename, pages, color=color,
            bitmapdpi=dpi, antialias=antialias,
            quality=quality, backcolor=backcolor,
            pdfdpi=pdfdpi, svgtextastext=svgtextastext)
        e.export()

    def Rename(self, widget, newname):
        """Rename the widget with the path given to the new name.

        eg Rename('graph1/xy1', 'scatter')
        This function does not move widgets."""

        w = self.document.resolveWidgetPath(self.currentwidget, widget)
        op = operations.OperationWidgetRename(w, newname)
        self.document.applyOperation(op)

    def NodeType(self, path):
        """This function treats the set of objects in the widget and
        setting tree as a set of nodes.

        Returns type of node given.
        Return values are: 'widget', 'settings' or 'setting'
        """
        item = self.document.resolvePath(self.currentwidget, path)

        if item.iswidget:
            return 'widget'
        elif item.issettings:
            return 'settinggroup'
        else:
            return 'setting'

    def NodeChildren(self, path, types='all'):
        """This function treats the set of objects in the widget and
        setting tree as a set of nodes.

        Returns a list of the names of the children of this node."""

        item = self.document.resolvePath(self.currentwidget, path)

        out = []
        if item.iswidget:
            if types == 'all' or types == 'widget':
                out += item.childnames
            if types == 'all' or types == 'settinggroup':
                out += [s.name for s in item.settings.getSettingsList()]
            if types == 'all' or types == 'setting':
                out += [s.name for s in item.settings.getSettingList()]
        elif item.issettings:
            if types == 'all' or types == 'settinggroup':
                out += [s.name for s in item.getSettingsList()]
            if types == 'all' or types == 'setting':
                out += [s.name for s in item.getSettingList()]
        return out

    def WidgetType(self, path):
        """Get the Veusz widget type for a widget with path given.

        Raises a ValueError if the path doesn't point to a widget."""

        item = self.document.resolvePath(self.currentwidget, path)
        if item.iswidget:
            return item.typename
        else:
            raise ValueError("Path '%s' is not a widget" % path)

    def SettingType(self, setting_path):
        """Get the type of setting (a string) for the path given.

        Raise a ValueError if path is not a setting
        """

        setn = self.document.resolveSettingPath(self.currentwidget, setting_path)
        return setn.typename

    def TagDatasets(self, tag, datasets):
        """Apply tag to list of datasets."""
        op = operations.OperationDataTag(tag, datasets)
        self.document.applyOperation(op)

        if self.verbose:
            print(_("Applied tag %s to datasets %s") % (
                tag, ' '.join(datasets)))

    def FilterDatasets(self, filterexpr, dataset_list,
                       prefix="", suffix="",
                       invert=False, replaceblanks=False):
        """Apply filter expression to list of datasets.

        filterexpr: input filter expression
        dataset_list: list of input dataset names
        prefix, suffix: output prefix/suffix to add to names (one must be set)
        invert: invert results of filter expression
        replaceblanks: replace filtered values with nan/blank in output.
        """
        op = operations.OperationDatasetsFilter(
            filterexpr, dataset_list,
            prefix=prefix, suffix=suffix,
            invert=invert, replaceblanks=replaceblanks)
        self.document.applyOperation(op)

        if self.verbose:
            print(
                _('Filtered datasets %s using expression %s. '
                  'Output prefix=%s, suffix=%s') % (
                      dataset_list, filterexpr, prefix, suffix)
            )
